@@ -28,13 +28,14 @@ from django.template import TemplateDoesNotExist
 from django.utils.decorators import method_decorator
 from django.views.decorators.csrf import csrf_exempt
 from django.views.generic import View
+from django.utils.module_loading import import_string
 from saml2 import BINDING_HTTP_POST, BINDING_HTTP_REDIRECT
 from saml2.client_base import LogoutError
 from saml2.config import SPConfig
 from saml2.ident import code, decode
 from saml2.mdstore import SourceNotFound
 from saml2.metadata import entity_descriptor
-from saml2.response import (RequestVersionTooLow, 
+from saml2.response import (RequestVersionTooLow,
                             SignatureError, StatusAuthnFailed, StatusError,
                             StatusNoAuthnContext, StatusRequestDenied,
                             UnsolicitedResponse)
@@ -83,7 +84,7 @@ class SPConfigMixin:
         conf = self.get_sp_config(request)
         state = StateCache(request.saml_session)
         client = Saml2Client(
-            conf, state_cache=state, 
+            conf, state_cache=state,
             identity_cache=IdentityCache(request.saml_session)
         )
         return state, client
@@ -103,7 +104,6 @@ class LoginView(SPConfigMixin, View):
         If set to None or nonexistent template, default form from the saml2 library
         will be rendered.
     """
-    logger.debug('Login process started')
 
     wayf_template = 'djangosaml2/wayf.html'
     authorization_error_template = 'djangosaml2/auth_error.html'
@@ -125,7 +125,6 @@ class LoginView(SPConfigMixin, View):
 
     def get(self, request, *args, **kwargs):
         logger.debug('Login process started')
-
         next_path = self.get_next_path(request)
 
         # if the user is already authenticated that maybe because of two reasons:
@@ -143,8 +142,8 @@ class LoginView(SPConfigMixin, View):
                 return HttpResponseRedirect(next_path)
             logger.debug('User is already logged in')
             return render(request, self.authorization_error_template, {
-                    'came_from': next_path,
-                    })
+                'came_from': next_path,
+            })
 
         try:
             conf = self.get_sp_config(request)
@@ -170,20 +169,22 @@ class LoginView(SPConfigMixin, View):
                                                   urlquote(came_from, safe=''))
                 ds_url = '{0}?entityID={1}&return={2}&returnIDParam=idp'
                 ds_url = ds_url.format(discovery_service,
-                                       urlquote(getattr(conf,'entityid'), safe=''),
+                                       urlquote(
+                                           getattr(conf, 'entityid'), safe=''),
                                        urlquote(login_url, safe=''))
                 return HttpResponseRedirect(ds_url)
 
             elif len(configured_idps) > 1:
                 logger.debug('A discovery process trough WAYF page is needed')
                 return render(request, self.wayf_template, {
-                        'available_idps': configured_idps.items(),
-                        'came_from': next_path,
-                        })
+                    'available_idps': configured_idps.items(),
+                    'came_from': next_path,
+                })
 
         # is the first one, otherwise next logger message will print None
         if not configured_idps:
-            raise IdPConfigurationMissing(('IdP configuration is missing or its metadata is expired.'))
+            raise IdPConfigurationMissing(
+                ('IdP configuration is missing or its metadata is expired.'))
         if selected_idp is None:
             selected_idp = list(configured_idps.keys())[0]
 
@@ -193,18 +194,23 @@ class LoginView(SPConfigMixin, View):
         logger.debug('Trying binding %s for IDP %s', binding, selected_idp)
 
         # ensure our selected binding is supported by the IDP
-        supported_bindings = get_idp_sso_supported_bindings(selected_idp, config=conf)
+        supported_bindings = get_idp_sso_supported_bindings(
+            selected_idp, config=conf)
         if binding not in supported_bindings:
-            logger.debug('Binding %s not in IDP %s supported bindings: %s', binding, selected_idp, supported_bindings)
+            logger.debug('Binding %s not in IDP %s supported bindings: %s',
+                         binding, selected_idp, supported_bindings)
             if binding == BINDING_HTTP_POST:
-                logger.warning('IDP %s does not support %s,  trying %s', selected_idp, binding, BINDING_HTTP_REDIRECT)
+                logger.warning('IDP %s does not support %s,  trying %s',
+                               selected_idp, binding, BINDING_HTTP_REDIRECT)
                 binding = BINDING_HTTP_REDIRECT
             else:
-                logger.warning('IDP %s does not support %s,  trying %s', selected_idp, binding, BINDING_HTTP_POST)
+                logger.warning('IDP %s does not support %s,  trying %s',
+                               selected_idp, binding, BINDING_HTTP_POST)
                 binding = BINDING_HTTP_POST
             # if switched binding still not supported, give up
             if binding not in supported_bindings:
-                raise UnsupportedBinding('IDP %s does not support %s or %s', selected_idp, BINDING_HTTP_POST, BINDING_HTTP_REDIRECT)
+                raise UnsupportedBinding(
+                    'IDP %s does not support %s or %s', selected_idp, BINDING_HTTP_POST, BINDING_HTTP_REDIRECT)
 
         client = Saml2Client(conf)
         http_response = None
@@ -223,8 +229,10 @@ class LoginView(SPConfigMixin, View):
                 if sign_requests:
                     # do not sign the xml itself, instead use the sigalg to
                     # generate the signature as a URL param
-                    sig_alg_option_map = {'sha1': SIG_RSA_SHA1, 'sha256': SIG_RSA_SHA256}
-                    sig_alg_option = getattr(conf, '_sp_authn_requests_signed_alg', 'sha1')
+                    sig_alg_option_map = {
+                        'sha1': SIG_RSA_SHA1, 'sha256': SIG_RSA_SHA256}
+                    sig_alg_option = getattr(
+                        conf, '_sp_authn_requests_signed_alg', 'sha1')
                     kwargs["sigalg"] = sig_alg_option_map[sig_alg_option]
                 session_id, result = client.prepare_for_authenticate(
                     entityid=selected_idp, relay_state=next_path,
@@ -251,15 +259,16 @@ class LoginView(SPConfigMixin, View):
                     if isinstance(request_xml, AuthnRequest):
                         # request_xml will be an instance of AuthnRequest if the message is not signed
                         request_xml = str(request_xml)
-                    saml_request = base64.b64encode(bytes(request_xml, 'UTF-8')).decode('utf-8')
+                    saml_request = base64.b64encode(
+                        bytes(request_xml, 'UTF-8')).decode('utf-8')
 
                     http_response = render(request, self.post_binding_form_template, {
                         'target_url': location,
                         'params': {
                             'SAMLRequest': saml_request,
                             'RelayState': next_path,
-                            },
-                        })
+                        },
+                    })
                 except TemplateDoesNotExist:
                     pass
 
@@ -280,7 +289,8 @@ class LoginView(SPConfigMixin, View):
         # success, so save the session ID and return our response
         oq_cache = OutstandingQueriesCache(request.saml_session)
         oq_cache.set(session_id, next_path)
-        logger.debug('Saving the session_id "%s" in the OutstandingQueries cache', oq_cache.__dict__)
+        logger.debug(
+            'Saving the session_id "%s" in the OutstandingQueries cache', oq_cache.__dict__)
         return http_response
 
 
@@ -292,14 +302,19 @@ class AssertionConsumerServiceView(SPConfigMixin, View):
         though some implementations may instead register their own subclasses of Saml2Backend.
     """
 
+    def custom_validation(self, response):
+        pass
+
     def handle_acs_failure(self, request, exception=None, status=403, **kwargs):
         """ Error handler if the login attempt fails. Override this to customize the error response.
         """
 
         # Backwards compatibility: if a custom setting was defined, use that one
-        custom_failure_function = get_custom_setting('SAML_ACS_FAILURE_RESPONSE_FUNCTION')
+        custom_failure_function = get_custom_setting(
+            'SAML_ACS_FAILURE_RESPONSE_FUNCTION')
         if custom_failure_function:
-            failure_function = custom_failure_function if callable(custom_failure_function) else import_string(custom_failure_function)
+            failure_function = custom_failure_function if callable(
+                custom_failure_function) else import_string(custom_failure_function)
             return failure_function(request, exception, status, **kwargs)
 
         return render(request, 'djangosaml2/login_error.html', {'exception': exception}, status=status)
@@ -312,11 +327,14 @@ class AssertionConsumerServiceView(SPConfigMixin, View):
             logger.warning('Missing "SAMLResponse" parameter in POST data.')
             return HttpResponseBadRequest('Missing "SAMLResponse" parameter in POST data.')
 
-        attribute_mapping = attribute_mapping or get_custom_setting('SAML_ATTRIBUTE_MAPPING', {'uid': ('username', )})
-        create_unknown_user = create_unknown_user or get_custom_setting('SAML_CREATE_UNKNOWN_USER', True)
+        attribute_mapping = attribute_mapping or get_custom_setting(
+            'SAML_ATTRIBUTE_MAPPING', {'uid': ('username', )})
+        create_unknown_user = create_unknown_user or get_custom_setting(
+            'SAML_CREATE_UNKNOWN_USER', True)
         conf = self.get_sp_config(request)
 
-        client = Saml2Client(conf, identity_cache=IdentityCache(request.saml_session))
+        identity_cache = IdentityCache(request.saml_session)
+        client = Saml2Client(conf, identity_cache=identity_cache)
         oq_cache = OutstandingQueriesCache(request.saml_session)
         oq_cache.sync()
         outstanding_queries = oq_cache.outstanding_queries()
@@ -331,38 +349,50 @@ class AssertionConsumerServiceView(SPConfigMixin, View):
             logger.exception("Error processing SAML Assertion.")
         except ResponseLifetimeExceed as e:
             _exception = e
-            logger.info(("SAML Assertion is no longer valid. Possibly caused by network delay or replay attack."), exc_info=True)
+            logger.info(
+                ("SAML Assertion is no longer valid. Possibly caused by network delay or replay attack."), exc_info=True)
         except SignatureError as e:
             _exception = e
             logger.info("Invalid or malformed SAML Assertion.", exc_info=True)
         except StatusAuthnFailed as e:
             _exception = e
-            logger.info("Authentication denied for user by IdP.", exc_info=True)
+            logger.info("Authentication denied for user by IdP.",
+                        exc_info=True)
         except StatusRequestDenied as e:
             _exception = e
             logger.warning("Authentication interrupted at IdP.", exc_info=True)
         except StatusNoAuthnContext as e:
             _exception = e
-            logger.warning("Missing Authentication Context from IdP.", exc_info=True)
+            logger.warning(
+                "Missing Authentication Context from IdP.", exc_info=True)
         except MissingKey as e:
             _exception = e
-            logger.exception("SAML Identity Provider is not configured correctly: certificate key is missing!")
+            logger.exception(
+                "SAML Identity Provider is not configured correctly: certificate key is missing!")
         except UnsolicitedResponse as e:
             _exception = e
-            logger.exception("Received SAMLResponse when no request has been made.")
+            logger.exception(
+                "Received SAMLResponse when no request has been made.")
         except RequestVersionTooLow as e:
             _exception = e
-            logger.exception("Received SAMLResponse have a deprecated SAML2 VERSION.")
+            logger.exception(
+                "Received SAMLResponse have a deprecated SAML2 VERSION.")
         except Exception as e:
             _exception = e
             logger.exception("SAMLResponse Error")
-        
+
         if _exception:
             return self.handle_acs_failure(request, exception=_exception)
         elif response is None:
             logger.warning("Invalid SAML Assertion received (unknown error).")
             return self.handle_acs_failure(request, status=400, exception=SuspiciousOperation('Unknown SAML2 error'))
 
+        try:
+            self.custom_validation(response)
+        except Exception as e:
+            logger.warning("SAML Response validation error: {e}")
+            return self.handle_acs_failure(request, status=400, exception=SuspiciousOperation('SAML2 validation error'))
+
         session_id = response.session_id()
         oq_cache.delete(session_id)
 
@@ -374,14 +404,17 @@ class AssertionConsumerServiceView(SPConfigMixin, View):
         if callable(create_unknown_user):
             create_unknown_user = create_unknown_user()
 
-        logger.debug('Trying to authenticate the user. Session info: %s', session_info)
+        logger.debug(
+            'Trying to authenticate the user. Session info: %s', session_info)
         user = auth.authenticate(request=request,
                                  session_info=session_info,
                                  attribute_mapping=attribute_mapping,
                                  create_unknown_user=create_unknown_user)
         if user is None:
-            logger.warning("Could not authenticate user received in SAML Assertion. Session info: %s", session_info)
-            return self.handle_acs_failure(request, exception=PermissionDenied('No user could be authenticated.'))
+            logger.warning(
+                "Could not authenticate user received in SAML Assertion. Session info: %s", session_info)
+            return self.handle_acs_failure(request, exception=PermissionDenied('No user could be authenticated.'),
+                                           session_info=session_info)
 
         auth.login(self.request, user)
         _set_subject_id(request.saml_session, session_info['name_id'])
@@ -391,7 +424,8 @@ class AssertionConsumerServiceView(SPConfigMixin, View):
         self.customize_session(user, session_info)
 
         relay_state = self.build_relay_state()
-        custom_redirect_url = self.custom_redirect(user, relay_state, session_info)
+        custom_redirect_url = self.custom_redirect(
+            user, relay_state, session_info)
         if custom_redirect_url:
             return HttpResponseRedirect(custom_redirect_url)
         relay_state = validate_referral_url(request, relay_state)
@@ -401,13 +435,14 @@ class AssertionConsumerServiceView(SPConfigMixin, View):
     def post_login_hook(self, request: HttpRequest, user: settings.AUTH_USER_MODEL, session_info: dict) -> None:
         """ If desired, a hook to add logic after a user has succesfully logged in.
         """
-        pass
 
     def build_relay_state(self) -> str:
         """ The relay state is a URL used to redirect the user to the view where they came from.
         """
-        default_relay_state = get_custom_setting('ACS_DEFAULT_REDIRECT_URL', settings.LOGIN_REDIRECT_URL)
-        relay_state = self.request.POST.get('RelayState', '/')
+        login_redirect_url = get_custom_setting('LOGIN_REDIRECT_URL', '/')
+        default_relay_state = get_custom_setting(
+            'ACS_DEFAULT_REDIRECT_URL', login_redirect_url)
+        relay_state = self.request.POST.get('RelayState', default_relay_state)
         relay_state = self.customize_relay_state(relay_state)
         if not relay_state:
             logger.warning('The RelayState parameter exists but is empty')
@@ -417,7 +452,6 @@ class AssertionConsumerServiceView(SPConfigMixin, View):
     def customize_session(self, user, session_info: dict):
         """ Subclasses can use this for customized functionality around user sessions.
         """
-        pass
 
     def customize_relay_state(self, relay_state: str) -> str:
         """ Subclasses may override this method to implement custom logic for relay state.
@@ -442,7 +476,8 @@ class EchoAttributesView(LoginRequiredMixin, SPConfigMixin, View):
 
         subject_id = _get_subject_id(request.saml_session)
         try:
-            identity = client.users.get_identity(subject_id, check_not_on_or_after=False)
+            identity = client.users.get_identity(
+                subject_id, check_not_on_or_after=False)
         except AttributeError:
             return HttpResponse("No active SAML identity found. Are you sure you have logged in via SAML?")
 
@@ -461,35 +496,41 @@ class LogoutInitView(LoginRequiredMixin, SPConfigMixin, View):
 
         subject_id = _get_subject_id(request.saml_session)
         if subject_id is None:
-            logger.warning('The session does not contain the subject id for user %s', request.user)
+            logger.warning(
+                'The session does not contain the subject id for user %s', request.user)
 
         try:
             result = client.global_logout(subject_id)
         except LogoutError as exp:
-            logger.exception('Error Handled - SLO not supported by IDP: {}'.format(exp))
+            logger.exception(
+                'Error Handled - SLO not supported by IDP: {}'.format(exp))
             auth.logout(request)
             state.sync()
-            return HttpResponseRedirect(getattr(settings, 'LOGOUT_REDIRECT_URL', '/'))
+            return self.handle_unsupported_slo_exception(request, exp)
 
         auth.logout(request)
         state.sync()
 
         if not result:
-            logger.error("Looks like the user %s is not logged in any IdP/AA", subject_id)
+            logger.error(
+                "Looks like the user %s is not logged in any IdP/AA", subject_id)
             return HttpResponseBadRequest("You are not logged in any IdP/AA")
 
         if len(result) > 1:
-            logger.error('Sorry, I do not know how to logout from several sources. I will logout just from the first one')
+            logger.error(
+                'Sorry, I do not know how to logout from several sources. I will logout just from the first one')
 
         for entityid, logout_info in result.items():
             if isinstance(logout_info, tuple):
                 binding, http_info = logout_info
                 if binding == BINDING_HTTP_POST:
-                    logger.debug('Returning form to the IdP to continue the logout process')
+                    logger.debug(
+                        'Returning form to the IdP to continue the logout process')
                     body = ''.join(http_info['data'])
                     return HttpResponse(body)
                 elif binding == BINDING_HTTP_REDIRECT:
-                    logger.debug('Redirecting to the IdP to continue the logout process')
+                    logger.debug(
+                        'Redirecting to the IdP to continue the logout process')
                     return HttpResponseRedirect(get_location(http_info))
                 else:
                     logger.error('Unknown binding: %s', binding)
@@ -497,9 +538,19 @@ class LogoutInitView(LoginRequiredMixin, SPConfigMixin, View):
             # We must have had a soap logout
             return finish_logout(request, logout_info)
 
-        logger.error('Could not logout because there only the HTTP_REDIRECT is supported')
+        logger.error(
+            'Could not logout because there only the HTTP_REDIRECT is supported')
         return HttpResponseServerError('Logout Binding not supported')
 
+    def handle_unsupported_slo_exception(self, request, exception, *args, **kwargs):
+        """ Subclasses may override this method to implement custom logic for
+            handling logout errors. Redirects to LOGOUT_REDIRECT_URL by default.
+
+            For example, a site may want to perform additional logic and redirect
+            users somewhere other than the LOGOUT_REDIRECT_URL.
+        """
+        return HttpResponseRedirect(getattr(settings, 'LOGOUT_REDIRECT_URL', '/'))
+
 
 @method_decorator(csrf_exempt, name='dispatch')
 class LogoutView(SPConfigMixin, View):
@@ -529,10 +580,12 @@ class LogoutView(SPConfigMixin, View):
         if 'SAMLResponse' in data:  # we started the logout
             logger.debug('Receiving a logout response from the IdP')
             try:
-                response = client.parse_logout_request_response(data['SAMLResponse'], binding)
+                response = client.parse_logout_request_response(
+                    data['SAMLResponse'], binding)
             except StatusError as e:
                 response = None
-                logger.warning("Error logging out from remote provider: " + str(e))
+                logger.warning(
+                    "Error logging out from remote provider: " + str(e))
             state.sync()
             return finish_logout(request, response)
 
@@ -572,7 +625,8 @@ def finish_logout(request, response, next_page=None):
     if (getattr(settings, 'SAML_IGNORE_LOGOUT_ERRORS', False) or (response and response.status_ok())):
         if not next_page:
             next_page = getattr(settings, 'LOGOUT_REDIRECT_URL', '/')
-        logger.debug('Performing django logout with a next_page of %s', next_page)
+        logger.debug(
+            'Performing django logout with a next_page of %s', next_page)
         return AuthLogoutView.as_view()(request, next_page=next_page)
     logger.error('Unknown error during the logout')
     return render(request, "djangosaml2/logout_error.html", {})
